/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.browser;

import javax.servlet.FilterChain;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import com.trifork.sosigw.Factory;
import com.trifork.sosigw.api.IDCardCache;
import com.trifork.sosigw.api.IDCardState;
import com.trifork.sosigw.cluster.IDCardClusterCacheEntry;
import com.trifork.sosigw.config.RuntimeConfig;
import com.trifork.sosigw.config.RuntimeConfigItem;
import com.trifork.sosigw.config.RuntimeConfigState;
import com.trifork.sosigw.config.StaticConfigApi;
import com.trifork.sosigw.config.StaticConfigImpl;
import com.trifork.sosigw.util.Util;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.*;

import java.util.Properties;

public class ConsoleLoginTest {
    
    private HttpServletRequest req;
    private HttpServletResponse res;
    private IDCardCache cache;
    private StaticConfigApi cfg;
    private FilterChain fc;

    @Before
    public void setup() {
        req = mock(HttpServletRequest.class);
        res = mock(HttpServletResponse.class);
        fc = mock(FilterChain.class);
        
        cache = mock(IDCardCache.class);
        Factory.setTheIDCardCache(cache);

        cfg = mock(StaticConfigApi.class);
        doReturn(new Properties()).when(cfg).getProperties();
        StaticConfigImpl.setInstance(cfg);
    }
    
    /**
     * Issue with Mockito and Thread Safety resulting in org.mockito.exceptions.misusing.WrongTypeOfReturnValue
     */
    @After
    public void tearDown() {
        StaticConfigImpl.setInstance(null);
    }
    
    @Test
    public void userNotLoggedInWhenNoCookies() throws Exception {
        assertFalse(ConsoleLogin.isUserLoggedIn(req));
        
        doReturn(new Cookie[] { new Cookie("name", "value") }).when(req).getCookies();
        assertFalse(ConsoleLogin.isUserLoggedIn(req));
    }
    
    @Test
    public void userLoggedInWhenIDCardIsSigned() throws Exception {
        doReturn(new Cookie[] { new Cookie(ConsoleLogin.CONSOLE_COOKIE, "test") }).when(req).getCookies();
        IDCardClusterCacheEntry entry = mock(IDCardClusterCacheEntry.class);
        doReturn(IDCardState.UNKNOWN).when(entry).getState();
        doReturn(entry).when(cache).findIDCardOrNull("test");
        
        assertFalse(ConsoleLogin.isUserLoggedIn(req));
        
        doReturn(IDCardState.SIGNED).when(entry).getState();
        assertTrue(ConsoleLogin.isUserLoggedIn(req));
    }
    
    @Test
    public void redirectWhenNotLoggedIn() throws Exception {
        doReturn(false).when(cfg).isBootstrapRestrictedMode();
        doReturn(false).when(cfg).isDeveloperRestrictedMode();
        
        ConsoleLogin cl = new ConsoleLogin();
        cl.doFilter(req, res, fc);
        verifyZeroInteractions(fc);
        verify(res).sendRedirect(anyString());
    }
    
    @Test
    public void allowAccessWhenBootstrap() throws Exception {
        doReturn(true).when(cfg).isBootstrapRestrictedMode();
        
        ConsoleLogin cl = new ConsoleLogin();
        cl.doFilter(req, res, fc);
        
        verify(fc).doFilter(req, res);
        assertNull(Util.getHttpServletRequest());
    }

    @Test
    public void allowAccessWhenDeveloper() throws Exception {
        doReturn(true).when(cfg).isDeveloperRestrictedMode();
        
        ConsoleLogin cl = new ConsoleLogin();
        cl.doFilter(req, res, fc);
        
        verify(fc).doFilter(req, res);
        assertNull(Util.getHttpServletRequest());
    }
    
    @Test
    public void testGetBrowserLoginUrl() throws Exception {
        Util.setActiveHttpRequest(req);
        IDCardClusterCacheEntry entry = new IDCardClusterCacheEntry("test");
        doReturn(entry).when(cache).ensureIDCard("test");
        doReturn("1.0.1").when(cfg).getSupportedDGWSVersion();
        assertNull(ConsoleLogin.getBrowserLoginUrl("1111111118", "test"));
        
        RuntimeConfig.setRuntimeConfigState(new RuntimeConfigState(1, new RuntimeConfigItem[] { new RuntimeConfigItem(RuntimeConfig.CATEGORY_CONSOLE_USERS, "1111111118", new String[] { "1", "2", "3" }) }));
        
        String url = ConsoleLogin.getBrowserLoginUrl("1111111118", "test");
        assertNotNull(url);

        verify(cache).entryUpdatedLocally(entry);
    }
}
