/*
* The MIT License
*
* Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
*
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* $HeadURL$
* $Id$
*/
package com.trifork.sosigw.ws;

import com.trifork.sosigw.Factory;
import com.trifork.sosigw.api.IDCardCache;
import com.trifork.sosigw.api.IDCardState;
import com.trifork.sosigw.cluster.IDCardClusterCacheEntry;
import com.trifork.sosigw.cluster.IDCardClusterCacheImpl;
import dk.sosi.gw._2007_09.DGWSFault;
import dk.sosi.gw._2007_09.SosiGWFacade;
import dk.sosi.seal.model.AuthenticationLevel;
import dk.sosi.seal.model.CareProvider;
import dk.sosi.seal.model.OIOBSTSAMLAssertionToIDCardResponse;
import dk.sosi.seal.model.SystemInfo;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.model.UserInfo;
import dk.sosi.seal.model.constants.DSTags;
import dk.sosi.seal.modelbuilders.OIOBSTSAMLAssertionToIDCardResponseModelBuilder;
import dk.sosi.seal.xml.XmlUtil;
import org.junit.Before;
import org.junit.Test;
import org.w3c.dom.Document;

import java.util.Properties;
import java.util.Scanner;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;


public abstract class CreateIdCardFromBSTTest extends AbstractIntegrationTests {

    protected abstract void setupStsSuccessResponse();
    protected abstract void setupStsErrorResponse();

    @Before
    public void setup() {
        Factory.setTheIDCardCache(new IDCardClusterCacheImpl(null));
    }

    /**
     * GW-T1
     * Prækondition: Bruger har ikke et idkort i GW-cachen i forvejen
     * Input: Validt createIdCardFromBST request uden angivelse af et ”sosi:SubjectNameID” claim
     * Output: Validt svar med det nyudstedte idkort i usigneret form
     * Sideeffekt: Det nyudstedte idkort er gemt i cachen under SubjectNameID fra bootstraptokenet og kan benyttes til at kalde en test-service (fx NTS) via GW
     */
    @Test
    public void testValidRequestWithoutNameIdClaimEmptyCache() throws Exception  {
        setupStsSuccessResponse();
        final IDCardCache idCardCache = Factory.getIDCardCache();
        assertEquals(0, idCardCache.size());
        SosiGWFacade sosiGWFacade = getFacade();
        String requestBody = getResourceString("OIOH3BST2SOSI-without-nameid-claim-request.xml");

        Document responseDoc = sendRequest(sosiGWFacade, requestBody);

        validateResponse(idCardCache, responseDoc, "KorsbaekKommune\\MSK");
    }

    /**
     * GW-T2
     * Prækondition: Bruger har ikke et idkort i GW-cachen i forvejen
     * Input: Validt createIdCardFromBST request med angivelse af et ”sosi:SubjectNameID” claim (som er forskelligt fra SubjectNameID i BST)
     * Output: Validt svar med det nyudstedte idkort i usigneret form
     * Sideeffekt: Det nyudstedte idkort er gemt i cachen under SubjectNameID fra claimet og kan benyttes til at kalde en test-service (fx NTS) via GW
     */
    @Test
    public void testValidRequestWithNameIdClaimEmptyCache() throws Exception  {
        setupStsSuccessResponse();
        final IDCardCache idCardCache = Factory.getIDCardCache();
        assertEquals(0, idCardCache.size());
        SosiGWFacade sosiGWFacade = getFacade();
        String requestBody = getResourceString("OIOH3BST2SOSI-with-nameid-claim-request.xml");

        Document responseDoc = sendRequest(sosiGWFacade, requestBody);

        validateResponse(idCardCache, responseDoc, "Mads_Skjern");
    }

    /**
     * GW-T3
     * Prækondition: Bruger har et idkort i GW-cachen i forvejen
     * Input: Validt createIdCardFromBST request uden angivelse af et ”sosi:SubjectNameID” claim
     * Output: Validt svar med det nyudstedte idkort i usigneret form
     * Sideeffekt: Det nyudstedte idkort er gemt i cachen under SubjectNameID fra bootstraptokenet, har erstattet det eksisterende idkort og kan benyttes til at kalde en test-service (fx NTS) via GW
     */
    @Test
    public void testValidRequestWithoutNameIdClaimWithIdCardInCache() throws Exception  {
        setupStsSuccessResponse();
        final IDCardCache idCardCache = Factory.getIDCardCache();
        saveTestEntryInCache(idCardCache, "KorsbaekKommune\\MSK");
        assertEquals(1, idCardCache.size());
        SosiGWFacade sosiGWFacade = getFacade();
        String requestBody = getResourceString("OIOH3BST2SOSI-without-nameid-claim-request.xml");

        Document responseDoc = sendRequest(sosiGWFacade, requestBody);

        validateResponse(idCardCache, responseDoc, "KorsbaekKommune\\MSK");
    }

    /**
     * GW-T4
     * Prækondition: Bruger har et idkort i GW-cachen i forvejen
     * Input: Validt createIdCardFromBST request med angivelse af et ”sosi:SubjectNameID” claim (som er forskelligt fra SubjectNameID i BST)
     * Output: Validt svar med det nyudstedte idkort i usigneret form
     * Sideeffekt: Det nyudstedte idkort er gemt i cachen under SubjectNameID fra claimet, har erstattet det eksisterende idkort og kan benyttes til at kalde en test-service (fx NTS) via GW
     */
    @Test
    public void testValidRequestWithNameIdClaimWithIdCardInCache() throws Exception  {
        setupStsSuccessResponse();
        final IDCardCache idCardCache = Factory.getIDCardCache();
        saveTestEntryInCache(idCardCache, "Mads_Skjern");
        assertEquals(1, idCardCache.size());
        SosiGWFacade sosiGWFacade = getFacade();
        String requestBody = getResourceString("OIOH3BST2SOSI-with-nameid-claim-request.xml");

        Document responseDoc = sendRequest(sosiGWFacade, requestBody);

        validateResponse(idCardCache, responseDoc, "Mads_Skjern");
    }

    /**
     * GW-T5
     * Prækondition: Bruger har et idkort i GW-cachen i forvejen
     * Input: Invalidt createIdCardFromBST request
     * Output: Fejlsvar uden idkort (videresendt i uforandret form fra STS’en)
     * Sideeffekt: Det eksisterende idkort slettes ikke fra cachen og kan fortsat benyttes til at kalde en test-service (fx NTS) via GW
     */
    @Test
    public void testInvalidRequestWithIdCardInCache() throws Exception  {
        setupStsErrorResponse();
        final IDCardCache idCardCache = Factory.getIDCardCache();
        IDCardClusterCacheEntry testCacheEntry = saveTestEntryInCache(idCardCache, "KorsbaekKommune\\MSK");
        assertEquals(1, idCardCache.size());
        SosiGWFacade sosiGWFacade = getFacade();
        String requestBody = "<Envelope/>";

        Document responseDoc = sendRequest(sosiGWFacade, requestBody);

        OIOBSTSAMLAssertionToIDCardResponse response = parseResponse(responseDoc);
        assertTrue(response.isFault());
        assertNull(response.getIDCard());
        validateCache(idCardCache, "KorsbaekKommune\\MSK", testCacheEntry.getIdCard(), testCacheEntry.getState());
    }

    /**
     * GW-T6
     * Prækondition: Bruger har ikke et idkort i GW-cachen i forvejen
     * Input: Invalidt createIdCardFromBST request
     * Output: Fejlsvar uden idkort (videresendt i uforandret form fra STS’en)
     * Sideeffekt: Ingen (brugeren har fortsat ikke nogen idkort i cachen)
     */
    @Test
    public void testInvalidRequestWithoutIdCardInCache() throws Exception  {
        setupStsErrorResponse();
        final IDCardCache idCardCache = Factory.getIDCardCache();
        assertEquals(0, idCardCache.size());
        SosiGWFacade sosiGWFacade = getFacade();
        String requestBody = "<Envelope/>";

        Document responseDoc = sendRequest(sosiGWFacade, requestBody);

        OIOBSTSAMLAssertionToIDCardResponse response = parseResponse(responseDoc);
        assertTrue(response.isFault());
        assertNull(response.getIDCard());
        assertEquals(0, idCardCache.size());
    }

    private IDCardClusterCacheEntry saveTestEntryInCache(IDCardCache idCardCache, String cacheKey) {
        IDCardClusterCacheEntry cachedIDCardEntry = idCardCache.ensureIDCard(cacheKey);
        final SystemInfo sysInfo = new SystemInfo(new CareProvider("medcom:skscode", "test", "test"), "test");
        final UserInfo userInfo = new UserInfo("1111111118", "Darkwing", "Duck", null, null, "role", null);
        UserIDCard userIDCard = new UserIDCard("1.0.1", AuthenticationLevel.MOCES_TRUSTED_USER, "test", sysInfo, userInfo, "e", "2", null, null);
        cachedIDCardEntry.setIdCard(userIDCard);
        idCardCache.updateEntryState(cachedIDCardEntry, IDCardState.REVOKED);
        idCardCache.entryUpdatedLocally(cachedIDCardEntry);
        return cachedIDCardEntry;
    }

    private void validateResponse(IDCardCache idCardCache, Document responseDoc, String cacheKey) {
        OIOBSTSAMLAssertionToIDCardResponse response = parseResponse(responseDoc);
        assertFalse(response.isFault());
        assertEquals("Response should not be signed", 0, responseDoc.getDocumentElement().getElementsByTagName(DSTags.SIGNATURE_PREFIXED).getLength());
        assertNotNull(response.getIDCard());
        assertTrue(response.getIDCard() instanceof  UserIDCard);
        UserIDCard idCard = (UserIDCard) response.getIDCard();

        validateCache(idCardCache, cacheKey, idCard, IDCardState.SIGNED);
    }

    private void validateCache(IDCardCache idCardCache, String cacheKey, UserIDCard expectedIdCard, IDCardState expectedState) {
        assertEquals(1, idCardCache.size());
        IDCardClusterCacheEntry cacheEntry = idCardCache.findIDCardOrNull(cacheKey);
        assertNotNull(cacheEntry);
        assertEquals(expectedState, cacheEntry.getState());
        assertEquals(expectedIdCard, cacheEntry.getIdCard());
    }

    private OIOBSTSAMLAssertionToIDCardResponse parseResponse(Document response) {
        return new OIOBSTSAMLAssertionToIDCardResponseModelBuilder().build(response);
    }

    private Document sendRequest(SosiGWFacade sosiGWFacade, String requestBody) throws DGWSFault {
        String responseString = sosiGWFacade.createIdCardFromBST(requestBody);
        assertNotNull(responseString);
        return XmlUtil.readXml(new Properties(), responseString, false);
    }

    String getResourceString(String resource) {
        return new Scanner(CreateIdCardFromBSTTest.class.getResourceAsStream(resource), "UTF-8").useDelimiter("\\A").next();
    }
}
